// MultiAperture_.java

import ij.*;
import ij.gui.*;
import ij.measure.*;
import ij.plugin.*;
import ij.process.*;
import ij.text.*;
import ij.gui.Toolbar.*; 

import java.awt.*;
import java.awt.event.*;
import java.awt.Canvas.*;
import java.util.*;

import astroj.*;


/**
 * Based on Aperture_.java, but with pre-selection of multiple apertures and processing of stacks.
 * 
 * @author F.V. Hessman, Georg-August-Universitaet Goettingen
 * @version 1.0
 * @date 2006-Feb-15
 *
 * @version 1.1
 * @date 2006-Nov-29
 * @changes Slight re-organization so that it's easier to use MultiAperture_ as a parent class, e.g. for Stack_Aligner_.
 *
 * @version 1.2
 * @date 2006-Dec-6
 * @changes Corrected problem with non-instantiated xOld for the case of images and not stacks;
 *	added thread to display current ROIs.
 *
 * @version 1.3
 * @date 2006-Dec-11
 * @changes Complicated thread solution replaced with simple overlay solution inherited from Aperture_
 *
 * @version 1.4
 * @date 2007-Mar-14
 * @changes Added <ESC> to stop processing of large stacks which have gone awry.
 *
 * @version 1.5
 * @date 2007-May-01
 * @changes Added output of ratio
 *
 * @version 1.6
 * @date 2010-Mar-18
 * @author Karen Collins (Univ. Louisville/KY)
 * @changes
 * 1) On my machine, the number of apertures and the 4 check box option values do not get stored between runs.
 *	I don't know if this is a problem with my computing environment, or if it is a problem in general.
 *	To make it work on my end, I had to implement the MultiAperture_ preferences retrieval from the Aperture_
 *	preferences management code. I was able to implement the save (set) from MultiAperture_ without problems.
 *	I could never work out how to do both from MultiAperture_, so my solution is probably not ideal, but it seems to work now.
 * 2) Implemented the Ratio Error functionality. The code I am currently using is:
 *		ratio*Math.sqrt(targetVariance/(target*target)+
 *		othersVariance/(others*others))        {equation 2}
 *	where
 *		ratio = target counts/total of all comparison counts  (all adjusted for sky background)
 *		targetVariance = (source error)^2 from the 1st of N apertures per image (as calculated in Photometer above)
 *		target = total source counts less sky background from the 1st of N apertures
 *		othersVariance = (ap#2 error)^2 + (ap#3 error)^2 + ...  + (ap#N error)^2
 *		others = total source counts less sky background from the 2nd through Nth apertures
 * 3) Implemented the Ratio SNR functionality. It simply reports (target counts)/{equation 2}, which is accomplished by
 *	not multiplying by "ratio".
 * 4) Measurements table code to support the ratio error and ratioSNR reporting.
 *
 * @version 1.7
 * @date 2011-Mar-29
 * @author F. Hessman (Goettingen)
 * @changes added dialog() to make it easier to sub-class (e.g. Stack_Aligner)
 */
public class MultiAperture_ extends Aperture_ implements MouseListener, KeyListener
	{
	boolean cancelled=false;
	boolean verbose=true;
	boolean blocked=false;
	boolean previous=false;
	boolean doStack=false;
    boolean processingStack=false;
    boolean apertureRemoved=false;
    boolean apertureChanged=false;
    boolean firstClick = true;
    int xLocation = 10, yLocation = 10;
	int firstSlice=1;
	int lastSlice=1;

    int astronomyToolId = 0;
    int currentToolId = 0;
    Toolbar toolbar;

	double[] xOld;
	double[] yOld;
    double[] ratio;
    double[] ratioError;
    double[] ratioSNR;
	double[] target;
	double[] targetVariance;

	String infoMessage = new String("");

//	double vx = 0.0;
//	double vy = 0.0;
//	double vxOld = 0.0;
//	double vyOld = 0.0;
//	public static String VX = new String ("vx");
//	public static String VY = new String ("vy");

	protected int ngot=0;
	protected int aperture=0;
    protected int nAperturesMax=500;
    protected int nApertures=2;
    protected int nAperturesStored=0;
    protected int startDragScreenX;
    protected int startDragScreenY;

	protected double[] xPos;
	protected double[] yPos;
	protected double[] xPosStored;
	protected double[] yPosStored;

	double[] others;		// SUM OF OTHER APERTURES
	double peak = 0.0;		// MAX PIXEL VALUE IN APERTURE
	double[] othersVariance;

    double apFWHMFactor = 1.4;
    double xFWHM = 0.0;
    double yFWHM = 0.0;

    double oldradius;
    double oldrBack1;
    double oldrBack2;

//	double ratio = 0.0;		// FIRST APERTURE
//	double ratioError = 0.0;
//	double ratioSNR = 0.0;
	public static String RATIO = new String ("rel_flux_T1");
    public static String OTHERS = new String ("tot_C_cnts");
    public static String OTHERS_ERROR = new String ("tot_C_err");

    protected boolean autoMode=false;
    protected boolean singleStep=false;
    protected boolean simulatedLeftClick=false;
    protected boolean aperturesInitialized=false;
    protected boolean allStoredAperturesPlaced=false;
    protected boolean enableDoubleClicks = true;
    protected boolean multiApertureRunning = false;

    protected boolean useVarSizeAp= false;
//	protected boolean follow=false;
//	protected boolean wideTable=true;

	protected boolean showRatio=true;
    protected boolean showCompTot=true;
	protected boolean showRatioError=true;
	protected boolean showRatioSNR=true;
    protected boolean updatePlot=true;
    protected boolean useMacroImage=false;
    protected String macroImageName=null;
    ImagePlus openImage;
    AstroStackWindow aw;
    

    protected boolean frameAdvance=false;

	String xOldApertures,yOldApertures;

    MouseEvent dummyClick;
    MouseEvent ee;

    protected int screenX;
    protected int screenY;
    protected int modis;

    TimerTask stackTask = null;
    java.util.Timer stackTaskTimer = null;

    boolean doubleClick = false;
    TimerTask doubleClickTask = null;
    java.util.Timer doubleClickTaskTimer = null;
	
    protected static String PREFS_AUTOMODE        = new String ("multiaperture.automode");  //0 click - for use with macros
    protected static String PREFS_FINISHED        = new String ("multiaperture.finished");  //signals finished to macros
    protected static String PREFS_CANCELED        = new String ("multiaperture.canceled");
    protected static String PREFS_PREVIOUS        = new String ("multiaperture.previous");
    protected static String PREFS_SINGLESTEP      = new String ("multiaperture.singlestep");
    protected static String PREFS_USEVARSIZEAP    = new String ("multiaperture.usevarsizeap");
    protected static String PREFS_APFWHMFACTOR    = new String ("multiaperture.apfwhmfactor");
//    protected static String PREFS_FOLLOW          = new String ("multiaperture.follow");
//	protected static String PREFS_WIDETABLE       = new String ("multiaperture.widetable");
	protected static String PREFS_SHOWRATIO       = new String ("multiaperture.showratio");
    protected static String PREFS_SHOWCOMPTOT     = new String ("multiaperture.showcomptot");
	protected static String PREFS_SHOWRATIO_ERROR = new String ("multiaperture.showratioerror");
	protected static String PREFS_SHOWRATIO_SNR   = new String ("multiaperture.showratiosnr");
    protected static String PREFS_NAPERTURESMAX   = new String ("multiaperture.naperturesmax");
	protected static String PREFS_XAPERTURES      = new String ("multiaperture.xapertures");
	protected static String PREFS_YAPERTURES      = new String ("multiaperture.yapertures");
    protected static String PREFS_USEMACROIMAGE   = new String ("multiaperture.useMacroImage");
    protected static String PREFS_MACROIMAGENAME  = new String ("multiaperture.macroImageName");
    protected static String PREFS_ENABLEDOUBLECLICKS  = new String ("multiaperture.enableDoubleClicks");
    protected static String PREFS_UPDATEPLOT  = new String ("multiaperture.updatePlot");
    protected static String PREFS_XLOCATION  = new String ("multiaperture.xLocation");
    protected static String PREFS_YLOCATION  = new String ("multiaperture.yLocation");
    protected static String PREFS_PREVNUMMONITORS  = new String ("multiaperture.prevNumberOfMonitors");

//	public static double RETRY_RADIUS = 3.0;



	/**
	 * Standard ImageJ PluginFilter setup routine.
	 */
	public int setup (String arg, ImagePlus imp)
		{
        this.getMeasurementPrefs();
        if (useMacroImage)
                {
                openImage = WindowManager.getImage(macroImageName);
                imp = openImage;
                }
        if (imp == null)
            {
            cancel();
            return DONE;
            }

        if (imp.getWindow() instanceof PlotWindow)     //if plotwindow is selected and only one other image window is open
            {                                          //switch to other image if not a plotwindow, otherwise message to user
            int numImages=WindowManager.getImageCount();
            if (numImages == 1)
                {
                IJ.showMessage("No image windows open for Multi-aperture processing");
                cancel();
                return DONE;
                }
            if (numImages == 2)
                {
                int[] idList = WindowManager.getIDList();
                imp = idList[0] == imp.getID() ? WindowManager.getImage(idList[1]) : WindowManager.getImage(idList[0]);
                if (imp.getWindow() instanceof PlotWindow)
                    {
                    IJ.showMessage("No image windows open for Multi-aperture processing");
                    cancel();
                    return DONE;
                    }
                }
            else
                {
                IJ.showMessage("Select image window to process, then restart Multi-aperture");
                cancel();
                return DONE;
                }
            }

        // TO MAKE SURE THAT THE NEXT CLICK WILL WORK
        toolbar = Toolbar.getInstance();
        astronomyToolId = toolbar.getToolId("Astronomy_Tool");
        currentToolId = Toolbar.getToolId();
        if (currentToolId != astronomyToolId)
            {
            if (astronomyToolId != -1)
            	IJ.setTool(astronomyToolId);
            else
                IJ.setTool(0);
            }

		IJ.register(MultiAperture_.class);
		return super.setup(arg,imp);
		}

	/**
	 * Standard ImageJ Plug-in method which runs the plug-in, notes whether the image is a stack,
	 * and registers the routine for mouse clicks.
	 */
	public void run (ImageProcessor ip)
		{
//        if (table==null)
//            IJ.log("Table is null");
//        else
//            IJ.log("Table rows = "+table.getCounter());
        Prefs.set (MultiAperture_.PREFS_CANCELED, "false");
		this.getMeasurementPrefs();
        T1="_T1";
        if (!autoMode)
            {
            String[] apsX = xOldApertures.split(",");
            String[] apsY = yOldApertures.split(",");
            if (apsX.length != apsY.length)
                nAperturesStored = 0;
            else
                nAperturesStored = apsX.length;
            }
        if (useMacroImage)
            {
            openImage = WindowManager.getImage(macroImageName);

            if (!(openImage == null))
                {
                imp = openImage;
                }
            }

		// GET HOW MANY APERTURES WILL BE MEASURED WITH WHAT RADII

		if ( !setUpApertures() || nApertures == 0 || !prepare () )
			{
            imp.unlock();
            cancel();
			return;
			}

		// START ESCAPE ABORTION POSSIBILITY
		IJ.resetEscape();

		// REGISTER FOR MOUSE CLICKS

        if ( !autoMode )
                {
                MouseListener[] ml = canvas.getMouseListeners();  
                for (int i = 0; i < ml.length; i++)
                    {
                    if (ml[i] instanceof MultiAperture_)
                        {
//                        IJ.log("MouseListener already running");
                        canvas.removeMouseListener(ml[i]);
                        }
                    }
                KeyListener[] kl = canvas.getKeyListeners();
                for (int i = 0; i < kl.length; i++)
                    {
                    if (kl[i] instanceof MultiAperture_)
                        {
//                        IJ.log("KeyListener already running");
                        canvas.removeKeyListener(kl[i]);
                        }
                    }
                canvas.addMouseListener(this);
                canvas.addKeyListener(this);
                }

		if (starOverlay || skyOverlay || valueOverlay || nameOverlay)
			ocanvas.clearRois();
        if (previous)
            {
            infoMessage = "Please select first aperture (right click to finalize) ...";
            IJ.showStatus (infoMessage);
            }
        setApertureColor(Color.green);
        setApertureName("T1");

        if (imp.getWindow() instanceof astroj.AstroStackWindow)
            {
            aw = (AstroStackWindow)imp.getWindow();
            aw.setDisableShiftClick(true);
            }
//        imp.getWindow().requestFocus();
//        imp.getCanvas().requestFocusInWindow();

		if ( autoMode )
            mouseReleased( dummyClick );
        else
            {
            imp.getWindow().requestFocus();
            imp.getCanvas().requestFocusInWindow();
            }
        }

    protected void cancel()
        {
        Prefs.set (MultiAperture_.PREFS_AUTOMODE, "false");
        Prefs.set (MultiAperture_.PREFS_FINISHED, "true");
        Prefs.set (MultiAperture_.PREFS_USEMACROIMAGE, "false");
        Prefs.set (MultiAperture_.PREFS_CANCELED, "true");
        }

	/**
	 * Get all preferences.
	 */
	protected void getMeasurementPrefs()
		{
		super.getMeasurementPrefs();

		autoMode       = Prefs.get (MultiAperture_.PREFS_AUTOMODE, autoMode);
        useMacroImage  = Prefs.get (MultiAperture_.PREFS_USEMACROIMAGE, useMacroImage);
        macroImageName = Prefs.get (MultiAperture_.PREFS_MACROIMAGENAME, macroImageName);
        previous       = Prefs.get (MultiAperture_.PREFS_PREVIOUS, previous);
        singleStep     = Prefs.get (MultiAperture_.PREFS_SINGLESTEP, singleStep);
//        wideTable      = Prefs.get (MultiAperture_.PREFS_WIDETABLE, wideTable);
//		follow         = false;	// Prefs.get (MultiAperture_.PREFS_FOLLOW,    follow);
        oldradius = radius;
        oldrBack1 = rBack1;
        oldrBack2 = rBack2;
		showRatio      = Prefs.get (MultiAperture_.PREFS_SHOWRATIO, showRatio);
        showCompTot    = Prefs.get (MultiAperture_.PREFS_SHOWCOMPTOT, showCompTot);
		showRatioError = Prefs.get (MultiAperture_.PREFS_SHOWRATIO_ERROR, showRatioError);
		showRatioSNR   = Prefs.get (MultiAperture_.PREFS_SHOWRATIO_SNR, showRatioSNR);
		useVarSizeAp   = Prefs.get (MultiAperture_.PREFS_USEVARSIZEAP, useVarSizeAp);
		apFWHMFactor   = Prefs.get (MultiAperture_.PREFS_APFWHMFACTOR, apFWHMFactor);
        nAperturesMax  = (int) Prefs.get (MultiAperture_.PREFS_NAPERTURESMAX, nAperturesMax);
        enableDoubleClicks = Prefs.get (MultiAperture_.PREFS_ENABLEDOUBLECLICKS, enableDoubleClicks);
		xOldApertures = Prefs.get (MultiAperture_.PREFS_XAPERTURES,"");
		yOldApertures = Prefs.get (MultiAperture_.PREFS_YAPERTURES,"");
        updatePlot = Prefs.get (MultiAperture_.PREFS_UPDATEPLOT,updatePlot);
        xLocation = (int) Prefs.get (MultiAperture_.PREFS_XLOCATION, xLocation);
        yLocation = (int) Prefs.get (MultiAperture_.PREFS_YLOCATION, yLocation);
		}

	/**
	 * Initializes variables etc.
	 */
	protected boolean prepare ()
		{
//		if (!checkResultsTable ()) return false;  //removed to allow dynamic number of apertures using right click to terminate aperture selections
		// LOAD PREVIOUS APERTURES IF DESIRED
		if ( previous )
			{
			String[] aps = xOldApertures.split(",");
			nAperturesStored = aps.length;
            xPosStored = extract(aps);
			aps = yOldApertures.split(",");
			if (nAperturesStored != aps.length)
				{
				IJ.error("The stored apertures are not consistent: "+nAperturesStored+"!="+aps.length);
				return false;
				}
            yPosStored = extract(aps);
			}
		if ( autoMode )
			{
			String[] aps = xOldApertures.split(",");
			nApertures = aps.length;
            xPos = extract(aps);
			aps = yOldApertures.split(",");
			if (nApertures != aps.length)
				{
				IJ.error("The stored apertures are not consistent: "+nApertures+"!="+aps.length);
				return false;
				}
            aperturesInitialized = true;
            yPos = extract(aps);
			}
		else	{
            xPos = new double[nApertures];
            yPos = new double[nApertures];
			}
		if (xPos == null || yPos == null)
			{
			IJ.error("Null aperture arrays???");
			return false;
			}

		imp.setSlice(firstSlice);
		ip = imp.getProcessor();
		imp.killRoi();
		return true;
		}

	/**
	 * Extracts a double array from a string array.
	 */
	protected double[] extract (String[] s)
		{
		double[] arr = new double[s.length];
		try	{
			for (int i=0; i < arr.length; i++)
				arr[i] = Double.parseDouble(s[i]);
			}
		catch (NumberFormatException e)
			{
			arr = null;
			}
		return arr;
		}

	/**
	 * Stops reception of mouse and keyboard clicks.
	 */
	protected void noMoreInput()
		{
		canvas.removeMouseListener(this);
        canvas.removeKeyListener(this);
        if (imp.getWindow() instanceof astroj.AstroStackWindow)
            {
            aw = (AstroStackWindow)imp.getWindow();
            aw.setDisableShiftClick(false);
            }
		}

	/**
	 * Finishes whole process.
	 */
	public void shutDown()
		{
        noMoreInput();
		super.shutDown();
        cancelled=true;
        processingStack=false;
        stackTask = null;
        stackTaskTimer = null;
        doubleClickTask = null;
        doubleClickTaskTimer = null;
        Prefs.set (MultiAperture_.PREFS_NAPERTURESMAX, nAperturesMax);
        Prefs.set (MultiAperture_.PREFS_PREVIOUS, previous);
        Prefs.set (MultiAperture_.PREFS_SINGLESTEP, singleStep);
        Prefs.set (MultiAperture_.PREFS_USEVARSIZEAP, useVarSizeAp);
        Prefs.set (MultiAperture_.PREFS_APFWHMFACTOR, apFWHMFactor);
//        Prefs.set (MultiAperture_.PREFS_WIDETABLE, wideTable);
        Prefs.set (MultiAperture_.PREFS_SHOWRATIO, showRatio);
        Prefs.set (MultiAperture_.PREFS_SHOWCOMPTOT, showCompTot);
        Prefs.set (MultiAperture_.PREFS_SHOWRATIO_ERROR, showRatioError);
        Prefs.set (MultiAperture_.PREFS_SHOWRATIO_SNR, showRatioSNR);
        Prefs.set (MultiAperture_.PREFS_ENABLEDOUBLECLICKS, enableDoubleClicks);
//        Prefs.set (MultiAperture_.PREFS_FOLLOW, follow);
        Prefs.set (MultiAperture_.PREFS_AUTOMODE, "false");
        Prefs.set (MultiAperture_.PREFS_FINISHED, "true");
        Prefs.set (MultiAperture_.PREFS_USEMACROIMAGE, "false");
        Prefs.set (MultiAperture_.PREFS_XLOCATION, xLocation);
        Prefs.set (MultiAperture_.PREFS_YLOCATION, yLocation);
		}

	//
	// MouseListener METHODS
	//

	/**
	 * Main MouseListener method used: process all mouse clicks.
	 */
	public void mouseReleased(MouseEvent e)
		{
        ee = e;
        if (autoMode || !enableDoubleClicks)
            {
            processSingleClick(ee);
            }
        else
            {
            if (e.getClickCount() == 1)
                {
                doubleClick = false;
                try	{
                    doubleClickTask = new TimerTask ()
                        {
                        public void run ()
                            {
                            if (!doubleClick) processSingleClick(ee);
                            doubleClickTask = null;
                            doubleClickTaskTimer = null;
                            }
                        };
                    doubleClickTaskTimer = new java.util.Timer();
                    if ((modis & InputEvent.BUTTON1_MASK) != 0)
                        doubleClickTaskTimer.schedule (doubleClickTask, 300);
                    else
                        doubleClickTaskTimer.schedule (doubleClickTask, 600);
                    }

                catch (Exception eee)
                    {
                    IJ.showMessage ("Error starting double click timer task : "+eee.getMessage());
                    }
                }
            else
                {
                doubleClick = true;
                return;
                }
            }
        }


    void processSingleClick(MouseEvent e)
        {
//        IJ.log("click count = "+e.getClickCount()+"    doubleClick = "+doubleClick);
        if (!autoMode)
            {
            screenX = e.getX();
            screenY = e.getY();
            modis = e.getModifiers();
            }

        //Right mouse click finalizes aperture selection or terminates single step mode
        if ((Math.abs(screenX-startDragScreenX) + Math.abs(screenY-startDragScreenY) < 4.0 ) &&
                (modis & InputEvent.BUTTON3_MASK) != 0)
            {
            if (!aperturesInitialized)
                {
                if (!previous || allStoredAperturesPlaced)
                    {
                    nApertures = ngot;
                    simulatedLeftClick = true;
                    aperturesInitialized = true;
                    }
                }
            else if (singleStep)
                {
                if (firstClick || allStoredAperturesPlaced)
                    {
                    IJ.beep();
                    shutDown();
                    }
                return;
                }
            }

        //do nothing unless automode or left mouse is clicked with no modifier keys pressed except "shift"
        if (autoMode || simulatedLeftClick || ((Math.abs(screenX-startDragScreenX) + Math.abs(screenY-startDragScreenY) < 4.0 ) &&
                (modis & InputEvent.BUTTON1_MASK) != 0 && 
                !e.isControlDown() && !e.isAltDown() && !e.isMetaDown()))
            {
            if (!autoMode && previous && !firstClick && !allStoredAperturesPlaced) return;
            simulatedLeftClick = false;
            if (ngot > 0)
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ngot+1));
                }
            else
                {
                setApertureColor(Color.green);
                setApertureName("T1");
                }
            if ( !autoMode && firstSlice > stackSize)
                {
                IJ.beep();
                shutDown();
                return;
                }
            if (frameAdvance)
                {
                firstSlice += 1;
                lastSlice = firstSlice;
                imp.setSlice(firstSlice);
                imp.updateImage();
                if (starOverlay || skyOverlay || valueOverlay || nameOverlay)
                    {
//                    ocanvas = OverlayCanvas.getOverlayCanvas (img);
//                    canvas = ocanvas;
                    ocanvas.clearRois();
                    }
                ip = imp.getProcessor();
                firstClick = true;
                allStoredAperturesPlaced = false;
                }
            else
                {
                slice = imp.getCurrentSlice();	

                // INITIAL CENTER FROM MOUSE CLICK OR PREVIOUS APERTURE LOCATION
                if (autoMode)
                    {
                    xCenter = xPos[0];
                    yCenter = yPos[0];
                    }
                else
                    {
                    xCenter = canvas.offScreenXD(e.getX());
                    yCenter = canvas.offScreenYD(e.getY());
                    if (ngot < nApertures && !(previous && firstClick))
                        {
                        apertureRemoved = ocanvas.removeRoi((int)xCenter, (int)yCenter);
                        if (!apertureRemoved && !measureAperture())
                            {
                            centerROI();
                            imp.unlock();
                            IJ.beep();
                            return;
                            }
                        }
                    }

                // ADD APERTURE TO LIST OR SHIFT OLD APERTURE POSITIONS

                if (autoMode)
                    {
                    if (!measureAperture())
                            {
                            centerROI();
                            imp.unlock();
                            IJ.beep();
                            return;
                            }
                    double dx = xCenter-xPos[0];
                    double dy = yCenter-yPos[0];
                    for (int i=0; i < nApertures; i++)
                        {
                        xPos[i] += dx;
                        yPos[i] += dy;
                        }
                    ngot = nApertures;
                    aperture = ngot;
                    }
                else if (previous && firstClick)
                    {
                    firstClick = false;
                    if (starOverlay || skyOverlay || valueOverlay || nameOverlay)
                        {
//                        ocanvas = OverlayCanvas.getOverlayCanvas (img);
//                        canvas = ocanvas;
                        ocanvas.clearRois();
                        }
                    double dx = xCenter-xPosStored[0];
                    double dy = yCenter-yPosStored[0];
                    for (int i=0; i < nAperturesStored; i++)
                        {
                        xPos[i] = xPosStored[i] + dx;
                        yPos[i] = yPosStored[i] + dy;
                        xCenter = xPos[i];
                        yCenter = yPos[i];
                        if (i > 0)
                            {
                            setApertureColor(Color.red);
                            setApertureName("C"+(i+1));
                            }
                        else
                            {
                            setApertureColor(Color.green);
                            setApertureName("T1");
                            }
                        if (!measureAperture())
                            {
                            centerROI();
                            ngot = 0;
                            setApertureColor(Color.green);
                            setApertureName("T1");
                            imp.unlock();
                            IJ.beep();
                            return;
                            }                        
                        }
                    ngot = nAperturesStored;
                    aperture = ngot;
                    allStoredAperturesPlaced = true;
                    }
                else if (!apertureRemoved && ngot < nApertures)
                    {
                    if (!e.isShiftDown() && ngot>0)
                        addAperture ();
                    else
                        addApertureAsT1();
                    }
                else if (apertureRemoved)
                    {
                    if (!e.isShiftDown())
                        removeAperture();
                    else
                        renameApertureToT1();
                    apertureRemoved = false;
                    }
                }

            if (frameAdvance)
                {
                if (firstSlice > stackSize)
                    IJ.showStatus ("Finished processing stack. Click to exit.");
                else
                    IJ.showStatus ("Click to select first aperture (right click to exit).");
                frameAdvance = false;
                }
            else if (singleStep && ngot >= nApertures)
                {
                //PROCESS ONE SLICE AT A TIME WHILE IN SINGLE STEP MODE
                xOld = (double[]) xPos.clone();
                yOld = (double[]) yPos.clone();
                if (!checkResultsTable())
                    {
                    IJ.showMessage("Multi-Aperture failed to create Measurements table");
                    IJ.beep();
                    shutDown();
                    }
                processStack();
                frameAdvance = true;
                saveNewApertures ();
                previous = true;
                Prefs.set (MultiAperture_.PREFS_PREVIOUS, previous);
                IJ.showStatus ("Click to advance slice (right click to exit).");
                }

            // GOT ALL APERTURES?
            else if (ngot < nApertures)
                {
                infoMessage = "Click to select aperture #"+(ngot+1)+" (<ESC> to abort).";
                IJ.showStatus (infoMessage);
                }
            else
                {	// PROCESS ALL SLICES WHEN NOT IN SINGLE STEP MODE
                noMoreInput ();
                xOld = (double[]) xPos.clone();
                yOld = (double[]) yPos.clone();
                saveNewApertures ();
                if (!checkResultsTable())
                    {
                    IJ.showMessage("Multi-Aperture failed to create Measurements table");
                    IJ.beep();
                    shutDown();
                    }
                if (!singleStep)
                    {
                    if (stackSize > 1 && doStack)
                        {
                        IJ.showStatus ("Processing stack...");
                        processingStack = true;
                        startProcessStack();
                        }
                    else
                        {
                        IJ.showStatus ("Processing image...");
                        processImage();
                        }
                    }
                if (!processingStack && !autoMode && (firstSlice < lastSlice)) IJ.beep();
                if (!processingStack) shutDown();
                }
            }
        }

        /**
	 * Saves new aperture locations to preferences.
	 */
        protected void saveNewApertures ()
            {
            String xpos = "";
            String ypos = "";
            for (int i=0; i < nApertures; i++)
                {
                if (i == 0)
                    {
                    xpos += (float)xPos[i];
                    ypos += (float)yPos[i];
                    }
                else
                    {
                    xpos += ","+(float)xPos[i];
                    ypos += ","+(float)yPos[i];
                    }
                }
            if (aperturesInitialized)
                {
                xPosStored = xPos.clone();
                yPosStored = yPos.clone();
                nAperturesStored = nApertures;
                Prefs.set (MultiAperture_.PREFS_XAPERTURES, xpos);
                Prefs.set (MultiAperture_.PREFS_YAPERTURES, ypos);
                }
            }


        /**
	 * Adds the aperture parameters to the list of apertures.
	 */
	protected void addAperture ()
		{
		xPos[ngot] = xCenter;
        yPos[ngot] = yCenter;
        ngot++;
		aperture = ngot;
		}

    protected void addApertureAsT1()
        {
        for (int ap = ngot-1; ap >= 0; ap--)
            {
            xPos[ap+1] = xPos[ap];
            yPos[ap+1] = yPos[ap];
            }
        xPos[0] = xCenter;
        yPos[0] = yCenter;
        ngot++;

        ocanvas.clearRois();
		for (int ap=0;  ap < ngot; ap++)
			{
			aperture = ap;
            if (ap == 0)
                {
                setApertureColor(Color.green);
                setApertureName("T1");
                }
            else
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ap+1));
                }

            xCenter = xPos[ap];
            yCenter = yPos[ap];
			if (!measureAperture())
				{
                centerROI();
                IJ.beep();
                IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while adding a different target aperture");
                shutDown();
                return;
				}
            }
        return;
        }

	protected void removeAperture ()
		{
        for (int ap = 0; ap < ngot; ap++)
            {
            if ((xCenter-xPos[ap])*(xCenter-xPos[ap])+(yCenter-yPos[ap])*(yCenter-yPos[ap]) <= radius*radius)
                {
                ngot--;
                for (int i = ap; i < ngot; i++)
                    {
                    xPos[i] = xPos[i+1];
                    yPos[i] = yPos[i+1];
                    }
                break;
                }
            }
        ocanvas.clearRois();
		for (int ap=0;  ap < ngot; ap++)
			{
			aperture = ap;
            if (ap == 0)
                {
                setApertureColor(Color.green);
                setApertureName("T1");
                }
            else
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ap+1));
                }

            xCenter = xPos[ap];
            yCenter = yPos[ap];
			if (!measureAperture())
				{
                centerROI();
                IJ.beep();
                IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while deleting a different aperture");
                shutDown();
                return;
				}
            }
        return;
		}

    protected void renameApertureToT1()
        {
        for (int ap = 0; ap < ngot; ap++)
            {
            if ((xCenter-xPos[ap])*(xCenter-xPos[ap])+(yCenter-yPos[ap])*(yCenter-yPos[ap]) <= radius*radius)
                {
                for (int i = ap; i > 0; i--)
                    {
                    xPos[i] = xPos[i-1];
                    yPos[i] = yPos[i-1];
                    }
                break;
                }
            }
        xPos[0] = xCenter;
        yPos[0] = yCenter;

        ocanvas.clearRois();
		for (int ap=0;  ap < ngot; ap++)
			{
			aperture = ap;
            if (ap == 0)
                {
                setApertureColor(Color.green);
                setApertureName("T1");
                }
            else
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ap+1));
                }

            xCenter = xPos[ap];
            yCenter = yPos[ap];
			if (!measureAperture())
				{
                centerROI();
                IJ.beep();
                IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while renaming an aperture to T1");
                shutDown();
                return;
				}
            }
        return;
        }


	public void mousePressed(MouseEvent e)
        {
        if (!autoMode)
            {
            startDragScreenX = e.getX();
            startDragScreenY = e.getY();
            }
        }
	public void mouseClicked(MouseEvent e) {}
	public void mouseExited(MouseEvent e)
			{
			IJ.showStatus (infoMessage);
			}
	public void mouseEntered(MouseEvent e) {}

        /** Handle the key typed event from the image canvas. */
    public void keyTyped(KeyEvent e) {

    }

    /** Handle the key-pressed event from the image canvas. */
    public void keyPressed(KeyEvent e) {
        int keyCode = e.getKeyCode();
        if (IJ.escapePressed())
            {
            IJ.beep();
            Prefs.set (MultiAperture_.PREFS_CANCELED, "true");
            shutDown();
            }
        else if (keyCode == KeyEvent.VK_DELETE)
            {
//            IJ.log("delete an aperture?");
            }
    }

    /** Handle the key-released event from the image canvas. */
    public void keyReleased(KeyEvent e) {

    }


	//
	// MultiAperture_ METHODS
	//

	/**
	 * Define the apertures and decide on the sub-stack if appropriate.
	 */
	protected boolean setUpApertures ()
		{
		// CHECK SLICES
        firstSlice = imp.getCurrentSlice();
		lastSlice = stackSize;
        if (singleStep)
                lastSlice = firstSlice;
        if (!autoMode)
                {
                GenericDialog gd = dialog();

                gd.showDialog();
                xLocation = gd.getX();
                yLocation = gd.getY();
                Prefs.set (MultiAperture_.PREFS_XLOCATION, xLocation);
                Prefs.set (MultiAperture_.PREFS_YLOCATION, yLocation);
                if (!gd.wasOKed())
                    {
                    cancelled = true;
                    return false;
                    }

			// GET UPDATED STANDARD PARAMETERS FROM REQUIRED DIALOG FIELDS:
			//	nApertures,firstSlice,lastSlice,previous,singleStep,oneTable,wideTable

			// NOTE: ONLY THE GENERIC MultiAperture_ FIELDS BELONG HERE !!!!!!!!!!!!!

                nAperturesMax = (int)gd.getNextNumber();
                nApertures = nAperturesMax;
                if (gd.invalidNumber() || nApertures <= 0)
                        {
                        IJ.beep();
                        IJ.error ("Invalid number of apertures: "+nApertures);
                        return false;
                        }
                Prefs.set (MultiAperture_.PREFS_NAPERTURESMAX, nAperturesMax);
                if (stackSize > 1)
                        {
                        firstSlice=(int)gd.getNextNumber();
                        if (gd.invalidNumber() || firstSlice < 1)
                                firstSlice=1;
                        lastSlice=(int)gd.getNextNumber();
                        if (gd.invalidNumber() || lastSlice > stackSize)
                                lastSlice= stackSize;
                        if (firstSlice != lastSlice)
                                {
                                if (firstSlice > lastSlice)
                                        {
                                        int i=firstSlice;
                                        firstSlice=lastSlice;
                                        lastSlice=i;
                                        }
                                doStack=true;
                                }
                        }
                else	{
                    firstSlice=1;
                    lastSlice=1;
                    }
                radius = gd.getNextNumber();
                if (gd.invalidNumber() || radius <= 1)
                        {
                        IJ.beep();
                        IJ.error ("Invalid aperture radius: "+radius);
                        return false;
                        }
                if (oldradius != radius)
                    changeRadii();
                rBack1 = gd.getNextNumber();
                if (gd.invalidNumber() || rBack1 < radius)
                        {
                        IJ.beep();
                        IJ.error ("Invalid background inner radius: "+rBack1);
                        return false;
                        }
                if (oldrBack1 != rBack1)
                    changeRadii();
                rBack2 = gd.getNextNumber();
                if (gd.invalidNumber() || rBack2 < rBack1)
                        {
                        IJ.beep();
                        IJ.error ("Invalid background outer radius: "+rBack2);
                        return false;
                        }
                if (oldrBack2 != rBack2)
                    changeRadii();
                previous = gd.getNextBoolean();
                singleStep = gd.getNextBoolean();
                if (singleStep)
                    lastSlice = firstSlice;
//                oneTable = !gd.getNextBoolean();
//                wideTable = gd.getNextBoolean();

			    Prefs.set (MultiAperture_.PREFS_PREVIOUS, previous);
			    Prefs.set (MultiAperture_.PREFS_SINGLESTEP, singleStep);
//                Prefs.set (MultiAperture_.PREFS_WIDETABLE, wideTable);

                // GET NON-STANDARD PARAMETERS AND CLEAN UP
			    return finishFancyDialog (gd);
                }

        return true;
		}

    protected void changeRadii()
        {
        apertureChanged = true;
        Prefs.set("setaperture.aperturechanged",apertureChanged);
		Prefs.set (Aperture_.AP_PREFS_RADIUS, radius);
		Prefs.set (Aperture_.AP_PREFS_RBACK1, rBack1);
		Prefs.set (Aperture_.AP_PREFS_RBACK2, rBack2);
        }


	protected void startProcessStack ()     //start new task to run stack processing so that table will update for each image processed
		{                                   //the mouse handler apparently blocks window updates to improve performance
		try	{
			stackTask = new TimerTask ()
				{
				public void run ()
					{
                    processStack();
                    stackTask = null;
                    stackTaskTimer = null;
                    }
                };
            stackTaskTimer = new java.util.Timer();
            stackTaskTimer.schedule (stackTask,0);
            }

        catch (Exception e)
            {
            IJ.showMessage ("Error starting process stack task : "+e.getMessage());
            }
        }


	/**
	 * Perform photometry on each image of selected sub-stack.
	 */
	protected void processStack ()
		{
		verbose=false;
//		vx = 0.0;
//		vy = 0.0;
		canvas = imp.getCanvas();
		ocanvas = null;
		for (int i=firstSlice; i <= lastSlice; i++)
			{
            slice=i;
			imp.setSlice(i);
			imp.updateImage();
			if (starOverlay || skyOverlay || valueOverlay || nameOverlay)
				{
				ocanvas = OverlayCanvas.getOverlayCanvas (imp);
				canvas = ocanvas;
				ocanvas.clearRois();
				}
			ip = imp.getProcessor();
            processImage();
			if (cancelled || IJ.escapePressed())
				{
				IJ.beep();
                Prefs.set (MultiAperture_.PREFS_CANCELED, "true");
                shutDown();
				return;
				}
			}
        if (processingStack)
            {
            IJ.beep();
            shutDown();
            return;
            }
        }

	/**
	 * Perform photometry on each aperture of current image.
	 */
	protected void processImage ()
		{
		double dx = 0.0;		// CHANGE
		double dy = 0.0;

		double src = 0.0;		// MEAN SOURCE BRIGHTNESSES AND BACKGROUNDS
		double bck = 0.0;

        ratio = new double[nApertures];
        ratioError = new double[nApertures];
        ratioSNR = new double[nApertures];
        target = new double[nApertures];
        targetVariance = new double[nApertures];
        others = new double[nApertures];
        othersVariance = new double[nApertures];
        
        for (int ap=0;  ap < nApertures; ap++)
            {      
            ratio[ap] = 0.0;
            ratioError[ap] = 0.0;
            ratioSNR[ap] = 0.0;
            target[ap] = 0.0;
            targetVariance[ap] = 0.0;
            others[ap] = 0.0;
            othersVariance[ap] = 0.0;
            }
        xFWHM = 0.0;
        yFWHM = 0.0;

        if (useVarSizeAp)
            {
            setVariableAperture(false);
            for (int ap=0;  ap < nApertures; ap++)
                {
                aperture = ap;
                // GET POSITION ESTIMATE
                if (ap == 0) setApertureColor(Color.green);
                else setApertureColor(Color.red);
//                if (follow)
//                    {
//                    xCenter=0.5*(xPos[ap]+xOld[ap]+vxOld)+vx;
//                    yCenter=0.5*(yPos[ap]+yOld[ap]+vyOld)+vy;
//                    }
//                else
//                    {
                    xCenter = xPos[ap];
                    yCenter = yPos[ap];
//                    }

                // MEASURE NEW POSITION
//                centerROI();
                if (!adjustAperture())
                    {
                    centerROI();
                    setVariableAperture(false);
                    IJ.beep();
                    IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while measuring FWHM");
                    shutDown();
                    return;
                    }

                xOld[ap] = xPos[ap];
                yOld[ap] = yPos[ap];

                xPos[ap] = xCenter;		// STORE POSITION IN CASE IT DRIFTS WITHIN A STACK
                yPos[ap] = yCenter;

                if (ap==0)
                    {
                    double xDel = xPos[0] - xOld[0];
                    double yDel = yPos[0] - yOld[0];
                    for (int app=1; app < nApertures; app++)
                        {
                        xOld[app] = xPos[app];
                        yOld[app] = yPos[app];

                        xPos[app] += xDel;
                        yPos[app] += yDel;
                        }
                    }

                xFWHM += xWidth;
                yFWHM += yWidth;
                }

            setVariableAperture(true, Math.max(xFWHM/nApertures, yFWHM/nApertures)*apFWHMFactor);
            }
        OverlayCanvas.getOverlayCanvas(imp).clearRois();
		for (int ap=0;  ap < nApertures; ap++)
			{
			aperture = ap;
            if (ap == 0)
                {
                setApertureColor(Color.green);
                setApertureName("T1");
                }
            else
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ap+1));
                }

			// GET POSITION ESTIMATE

//			if (follow)
//				{
//				xCenter=0.5*(xPos[ap]+xOld[ap]+vxOld)+vx;
//				yCenter=0.5*(yPos[ap]+yOld[ap]+vyOld)+vy;
//				}
//			else
//				{
				xCenter = xPos[ap];
				yCenter = yPos[ap];
//				}

			// MEASURE NEW POSITION

//			centerROI();

			if (!measureAperture())
				{
                centerROI();
                setVariableAperture(false);
                IJ.beep();
                IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while measuring photometric values");
                shutDown();
                return;
				}

            // STORE RESULTS

			if (ap == 0)
				{
			 	storeResults();
				}
            else
				storeAdditionalResults (ap);

            // FOLLOW MOTION FROM FRAME TO FRAME

			dx += xCenter-xPos[ap];
			dy += yCenter-yPos[ap];

			xOld[ap] = xPos[ap];
			yOld[ap] = yPos[ap];

			xPos[ap]=xCenter;		// STORE POSITION IN CASE IT DRIFTS WITHIN A STACK
			yPos[ap]=yCenter;

			src += source;
			bck += back;

            ratio[ap] = source;
            target[ap] = source;
            targetVariance[ap] = serror*serror;
            if (ap != 0)
                {
                others[0]+= source;
                othersVariance[0] += serror*serror;
                for (int i=1; i<nApertures; i++)
                    {
                    if (i != ap)
                        {
                        others[i] += source;
                        othersVariance[i] += serror*serror;
                        }
                    }
                }
            else
                {
                double xDel = xPos[0] - xOld[0];
                double yDel = yPos[0] - yOld[0];
                for (int app=1; app < nApertures; app++)
                    {
                    xOld[app] = xPos[app];
                    yOld[app] = yPos[app];

                    xPos[app] += xDel;
                    yPos[app] += yDel;
                    }
                }

			// FOR DAUGHTER CLASSES....

			noteOtherApertureProperty (ap);
			}


        setVariableAperture(false);

		// COMPUTE APERTURE RATIO AND ERRORS AND UPDATE TABLE
        if (!isInstanceOfStackAlign)
            {
            if (showRatio && nApertures > 1)
                {
                for (int ap=0;  ap < (nApertures==2?1:nApertures); ap++)
                    {
                    ratio[ap] /= others[ap];
                    table.addValue ("rel_flux_"+(ap==0?"T":"C")+(ap+1), ratio[ap], 6);
                    if (showRatioError)
                        table.addValue ("rel_flux_err_"+(ap==0?"T":"C")+(ap+1), ratio[ap]*Math.sqrt(targetVariance[ap]/(target[ap]*target[ap])+ othersVariance[ap]/(others[ap]*others[ap])), 6);
                    if (showRatioSNR)
                        table.addValue ("rel_flux_SNR_"+(ap==0?"T":"C")+(ap+1), 1/Math.sqrt(targetVariance[ap]/(target[ap]*target[ap])+ othersVariance[ap]/(others[ap]*others[ap])), 6);
                    }
                }
            if (showCompTot && nApertures > 2)
                {
                for (int ap=0; ap<nApertures; ap++)
                    {
                    table.addValue (OTHERS+(ap==0?"":"-C"+(ap+1)), others[ap], 6);
                    }
                }
            if (showCompTot && showErrors && nApertures > 2)
                {
                for (int ap=0; ap<nApertures; ap++)
                    {
                    table.addValue (OTHERS_ERROR+(ap==0?"":"-C"+(ap+1)), Math.sqrt(othersVariance[ap]), 6);
                    }
                }
            }
            
		// CALCULATE MEAN SHIFT, BRIGHTNESS, AND BACKGROUND

		xCenter = dx/nApertures;
		yCenter = dy/nApertures;
		source = src/nApertures;
		back = bck/nApertures;

		// CALCULATE AND NOTE THE PIXEL VELOCITIES IN PIXELS/SLICE

//		if (follow)
//			{
//			vxOld = vx;
//			vyOld = vy;
//			vx = dx/nApertures;
//			vy = dy/nApertures;
//			table.addValue (VX, vx, 6);
//			table.addValue (VY, vy, 6);
//			}


		// UPDATE TABLE
        if (!isInstanceOfStackAlign)
            {        
            table.show();
            tablePanel = MeasurementTable.getTextPanel(tableName);
            if (tablePanel!=null)
                {int lastLine = tablePanel.getLineCount()-1;
                tablePanel.setSelection(lastLine, lastLine);
                tablePanel.resetSelection();
                }
            if ((updatePlot && !Data_Processor.active) || (Data_Processor.active && Data_Processor.runMultiPlot))  
                {
                if (MultiPlot_.mainFrame!=null)
                    {
                    if (MultiPlot_.getTable()!=null && MultiPlot_.getTable().equals(table))
                        {
//                        IJ.log("update plot");
                        MultiPlot_.updatePlot();
                        }
                    else
                        {
//                        IJ.log("setTable");
                        MultiPlot_.setTable(table, false);
                        }
                    }
                else
                    {
                    IJ.runPlugIn("MultiPlot_",tableName);
                    if (MultiPlot_.mainFrame!=null && MultiPlot_.getTable()!=null)
                        {
//                        IJ.log("setTable first time");
                        MultiPlot_.setTable(table, false);
                        }
                    }
                }
            }
        }
    
    

	/**
	 * Stores results for additional apertures.
	 */
	void storeAdditionalResults (int ap)
		{
		if (isInstanceOfStackAlign || ap <= 0) return;

		String header = "_"+(ap==0?"T":"C")+(ap+1);
		if (showPosition)
			{
			table.addValue (AP_XCENTER+header, xCenter, 6);
			table.addValue (AP_YCENTER+header, yCenter, 6);
			}
		if (showPositionFITS)
			{
			table.addValue (AP_XCENTER_FITS+header, xCenter + Centroid.PIXELCENTER, 6);
			table.addValue (AP_YCENTER_FITS+header, (double)imp.getHeight() - yCenter + Centroid.PIXELCENTER, 6);
			}        
		if (showRADEC && wcs.hasRaDec() && raDec != null)
			{
			table.addValue (AP_RA+header, raDec[0]/15.0, 6);
			table.addValue (AP_DEC+header, raDec[1], 6);
			}
		if (showPhotometry)
			{
			table.addValue (AP_SOURCE+header,   photom.sourceBrightness(), 6);
			if (showErrors) table.addValue (AP_SOURCE_ERROR+header,   photom.sourceError(), 6);
            if (showSNR) table.addValue (AP_SOURCE_SNR+header,   photom.sourceBrightness()/photom.sourceError(), 6);
			if (showBack) table.addValue (AP_BACK+header, photom.backgroundBrightness());
		  	if (showPeak) table.addValue (AP_PEAK+header, photom.peakBrightness(), 6);
			if (showSaturationWarning && photom.peakBrightness() > saturationWarningLevel &&
                photom.peakBrightness() > table.getValue(AP_WARNING, table.getCounter()-1))
                    {
                    table.setValue(AP_WARNING, table.getCounter()-1, photom.peakBrightness());
                    }
			}
		if (showWidths)
			{
			table.addValue (AP_XWIDTH+header,   xWidth, 6);
			table.addValue (AP_YWIDTH+header,   yWidth, 6);
			}
		if (showMeanWidth)
			table.addValue (AP_MEANWIDTH+header, 0.5*(xWidth+yWidth), 6);
//		table.show();
		}


	/**
	 * Notes anything else which might be interesting about an aperture measurement.
	 */
	protected void noteOtherApertureProperty  (int ap)
		{
        if (isInstanceOfStackAlign) return;
		}
    
    static public void clearTable()
        {
        if (table != null && MeasurementTable.getTextPanel(tableName) != null)
            {
            table = new MeasurementTable(tableName);
            table.show();
            }
        else
            {
            IJ.showStatus("No table to clear");
            }
        }

	/**
	 * Set up extended table format.
	 */
	protected boolean checkResultsTable ()
		{
        if (isInstanceOfStackAlign) return true;
        
        tablePanel = MeasurementTable.getTextPanel(tableName);
        
        if (table != null && tablePanel != null && table.getCounter() > 0) return true;
        
//		if (!super.checkResultsTable ()) return false;
        
	    table = MeasurementTable.getTable (tableName);

		if (table == null)
			{
			IJ.error ("Unable to open measurement table.");
			return false;
			}        
        
		int i=0;

        tablePanel = MeasurementTable.getTextPanel(tableName);
        
        if (showSliceNumber && table.getColumnIndex(AP_SLICE) == ResultsTable.COLUMN_NOT_FOUND)
            i=table.getFreeColumn (AP_SLICE);
        if (showSaturationWarning && table.getColumnIndex(AP_WARNING) == ResultsTable.COLUMN_NOT_FOUND)
            i=table.getFreeColumn (AP_WARNING);
		if (showTimes && table.getColumnIndex(AP_MJD) == ResultsTable.COLUMN_NOT_FOUND)
			i=table.getFreeColumn (AP_MJD);
		if (showFits && fitsKeywords != null)
			{
			String[] sarr = fitsKeywords.split(",");
			for (int l=0; l < sarr.length; l++)
				{
				if (!sarr[l].equals("") &&
						table.getColumnIndex(sarr[l]) == ResultsTable.COLUMN_NOT_FOUND)
					i=table.getFreeColumn (sarr[l]);
				}
			}
        if (showRadii)
            {
			if (table.getColumnIndex(AP_RSOURCE) == ResultsTable.COLUMN_NOT_FOUND)
				i=table.getFreeColumn (AP_RSOURCE);
			if (table.getColumnIndex(AP_RBACK1) == ResultsTable.COLUMN_NOT_FOUND)
				i=table.getFreeColumn (AP_RBACK1);
			if (table.getColumnIndex(AP_RBACK2) == ResultsTable.COLUMN_NOT_FOUND)
				i=table.getFreeColumn (AP_RBACK2);
            }


		if (nApertures > 1)
			{
			if (showRatio)
				{
                for (int ap=1; ap <= nApertures; ap++)
                    {
                    if (table.getColumnIndex("rel_flux_"+(ap==1?"T":"C")+ap) == MeasurementTable.COLUMN_NOT_FOUND)
                        i=table.getFreeColumn ("rel_flux_"+(ap==1?"T":"C")+ap);
                    }
				if (showRatioError)
					{
                    for (int ap=1; ap <= nApertures; ap++)
                        {
                        if (table.getColumnIndex("rel_flux_err_"+(ap==1?"T":"C")+ap) == MeasurementTable.COLUMN_NOT_FOUND)
                            i=table.getFreeColumn ("rel_flux_err_"+(ap==1?"T":"C")+ap);
                        }
                    }
				if (showRatioSNR)
					{
                    for (int ap=1; ap <= nApertures; ap++)
                        {
                        if (table.getColumnIndex("rel_flux_SNR_"+(ap==1?"T":"C")+ap) == MeasurementTable.COLUMN_NOT_FOUND)
                            i=table.getFreeColumn ("rel_flux_SNR_"+(ap==1?"T":"C")+ap);
                        }
                    }
				}
            if (showCompTot && nApertures>2)
                {
                for (int ap=0; ap<nApertures; ap++)
                    {
                    if (table.getColumnIndex(OTHERS+(ap==0?"":"-C"+(ap+1))) == MeasurementTable.COLUMN_NOT_FOUND)
                        i=table.getFreeColumn (OTHERS+(ap==0?"":"-C"+(ap+1)));
                    if (showErrors && table.getColumnIndex(OTHERS_ERROR+(ap==0?"":"-C"+(ap+1))) == MeasurementTable.COLUMN_NOT_FOUND)
                        i=table.getFreeColumn (OTHERS_ERROR+(ap==0?"":"-C"+(ap+1)));
                    }
                }

            if (showPosition && table.getColumnIndex(AP_XCENTER+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_XCENTER+T1);
            if (showPosition && table.getColumnIndex(AP_YCENTER+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_YCENTER+T1);
            if (showPositionFITS && table.getColumnIndex(AP_XCENTER_FITS+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_XCENTER_FITS+T1);
            if (showPositionFITS && table.getColumnIndex(AP_YCENTER_FITS+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_YCENTER_FITS+T1);            
            if (showRADEC && wcs.hasRaDec() && raDec != null)
                {
                if (table.getColumnIndex(AP_RA+T1) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_RA+T1);
                if (table.getColumnIndex(AP_DEC+T1) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_DEC+T1);
                }
            if (showPhotometry && table.getColumnIndex(AP_SOURCE+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_SOURCE+T1);
            if (showErrors && table.getColumnIndex(AP_SOURCE_ERROR+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_SOURCE_ERROR+T1);
            if (showSNR && table.getColumnIndex(AP_SOURCE_SNR+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_SOURCE_SNR+T1);
            if (showPeak && table.getColumnIndex(AP_PEAK+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_PEAK+T1);
            if (showBack && table.getColumnIndex(AP_BACK+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_BACK+T1);
            if (showMeanWidth && table.getColumnIndex(AP_MEANWIDTH+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_MEANWIDTH+T1);
            if (showWidths)
                {
                if (table.getColumnIndex(AP_XWIDTH+T1) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_XWIDTH+T1);
                if (table.getColumnIndex(AP_YWIDTH+T1) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_YWIDTH+T1);
                }
            if (showAngle && table.getColumnIndex(AP_ANGLE+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_ANGLE+T1);
            if (showRoundness && table.getColumnIndex(AP_ROUNDNESS+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_ROUNDNESS+T1);
            if (showVariance && table.getColumnIndex(AP_VARIANCE+T1) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_VARIANCE+T1);
            if (showRaw)
                {
                if (table.getColumnIndex(AP_RSOURCE) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_RSOURCE);
                if (table.getColumnIndex(AP_RBACK1) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_RBACK1);
                if (table.getColumnIndex(AP_RBACK2) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_RBACK2);
                if (isCalibrated)
                    {
                    if (table.getColumnIndex(AP_RAWSOURCE+T1) == ResultsTable.COLUMN_NOT_FOUND)
                        i=table.getFreeColumn (AP_RAWSOURCE+T1);
                    if (table.getColumnIndex(AP_RAWBACK+T1) == ResultsTable.COLUMN_NOT_FOUND)
                        i=table.getFreeColumn (AP_RAWBACK+T1);
                    }
                }            
            
            for (int ap=1; ap < nApertures; ap++)
				{
				String header = "_C"+(ap+1);
				if (showPosition)
					{
					if (table.getColumnIndex( AP_XCENTER+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_XCENTER+header);
					if (table.getColumnIndex( AP_YCENTER+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_YCENTER+header);
					}
				if (showPositionFITS)
					{
					if (table.getColumnIndex( AP_XCENTER_FITS+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_XCENTER_FITS+header);
					if (table.getColumnIndex( AP_YCENTER_FITS+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_YCENTER_FITS+header);
					}                
                if (showRADEC && wcs.hasRaDec() && raDec != null)
                    {
                    if (table.getColumnIndex(AP_RA+header) == ResultsTable.COLUMN_NOT_FOUND)
                        i=table.getFreeColumn (AP_RA+header);
                    if (table.getColumnIndex(AP_DEC+header) == ResultsTable.COLUMN_NOT_FOUND)
                        i=table.getFreeColumn (AP_DEC+header);
                    }
				if (showPhotometry)
					{
					if (table.getColumnIndex( AP_SOURCE+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_SOURCE+header);
					if (showErrors && table.getColumnIndex(AP_SOURCE_ERROR+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_SOURCE_ERROR+header);
					if (showSNR && table.getColumnIndex(AP_SOURCE_SNR+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_SOURCE_SNR+header);
					if (showPeak && table.getColumnIndex(     AP_PEAK+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_PEAK+header);
					if (table.getColumnIndex( AP_BACK+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_BACK+header);
					}
				if (showMeanWidth && table.getColumnIndex ( AP_MEANWIDTH) == MeasurementTable.COLUMN_NOT_FOUND)
					i=table.getFreeColumn (AP_MEANWIDTH+header);
				if (showWidths)
					{
					if (table.getColumnIndex( AP_XWIDTH+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_XWIDTH+header);
					if (table.getColumnIndex( AP_YWIDTH+header) == MeasurementTable.COLUMN_NOT_FOUND)
						i=table.getFreeColumn ( AP_YWIDTH+header);
					}
				}
			}
		table.show();
		return true;
		}


	/**
	 * Add aperture number to overlay display.
	 */
	protected void addStringRoi (double x, double y, String text)
		{
		super.addStringRoi (x,y,"  #"+(aperture+1)+":  "+text.trim());
		}

	/**
	 * Standard preferences dialog for MultiAperture_
	 */
	protected GenericDialog dialog()
		{
        if (!Prefs.isLocationOnScreen(new Point(xLocation, yLocation)))
            {
            xLocation = 10;
            yLocation = 10;
            }

		GenericDialog gd = new GenericDialog("Multi-Aperture Measurements", xLocation, yLocation);
//		gd.addMessage ("Aperture radii should have been set with the \"Set Aperture\" tool (double-click icon).");

		gd.addNumericField ("   Maximum number of apertures per image :", nAperturesMax,0,6,"  (right click in image to finalize)");
        gd.addMessage("");
		if (stackSize > 1)
			{
            gd.addSlider("           First slice ", 1, stackSize, firstSlice == stackSize ? 1 : firstSlice);
            gd.addSlider("           Last slice ", 1, stackSize, lastSlice);
//			gd.addNumericField ("           First slice :", firstSlice == stackSize ? 1 : firstSlice,0);
//			gd.addNumericField ("           Last slice :",  lastSlice, 0);
	        }
        gd.addMessage("");
        gd.addSlider("Radius of object aperture", 1, radius>100?radius:100, radius);
        gd.addSlider("Inner radius of background annulus", 1, rBack1>100?rBack1:100, rBack1);
        gd.addSlider("Outer radius of background annulus", 1, rBack2>100?rBack2:100, rBack2);
//        gd.addNumericField ("Radius of object aperture",radius,2);
//		gd.addNumericField ("Inner radius of background annulus",rBack1,2);
//		gd.addNumericField ("Outer radius of background annulus",rBack2,2);
        gd.addMessage("");
		gd.addCheckbox ("Use previous "+nAperturesStored+" apertures (1-click to set first aperture location).",previous && nAperturesStored > 0);
		gd.addCheckbox ("Use single step mode (set new aperture locations for each image - right click to exit)",singleStep);
//		gd.addCheckbox ("Put results in image's own measurements table.", !oneTable);
//		gd.addCheckbox ("All measurements from one image on the same Measurements Table line.", wideTable);
		gd.addMessage (" ");

		// HERE ARE THE THINGS WHICH AREN'T ABSOLUTELY NECESSARY
		addFancyDialog (gd);

		// gd.addCheckbox ("Track moving object (experimental!).",follow);
		// gd.addMessage (" ");
		gd.addMessage ("PRESS \"OK\" AND SELECT APERTURE LOCATIONS WITH LEFT CLICKS. THEN RIGHT CLICK TO BEGIN PROCESSING.");
        gd.addMessage ("(to abort aperture selection or processing, press <ESC>)");
		return gd;
		}

	/**
	 * Adds options to MultiAperture_ dialog() which aren't absolutely necessary.
	 * Sub-classes of MultiAperture_ may choose to replace or extend this functionality if they use the original dialog().
	 */
	protected void addFancyDialog (GenericDialog gd)
		{
		// GET NON-REQUIRED DIALOGUE FIELDS:
		//	showRatio,showRatioError,showRatioSNR,useVarSizeAp,apFWHMFactor
        gd.addCheckbox ("Reposition aperture to object centroid (leave unchecked to center aperture at mouse click).", reposition);
        gd.addCheckbox ("Update plot of measurements while running.", updatePlot);
		gd.addCheckbox ("Compute relative flux in all apertures (T1/Sum(CX) and CN/Sum(CX) where X not equal N)",showRatio);
		gd.addCheckbox ("Compute relative flux error (only if you check \"Compute relative flux\" above).",showRatioError);
		gd.addCheckbox ("Compute relative flux signal-to-noise (only if you check \"Compute relative flux\" above).",showRatioSNR);
        gd.addCheckbox ("Compute total comparison star counts (for apertures 2 to N).",showCompTot);
		gd.addCheckbox ("Vary photometer aperture radii based on FWHM.",useVarSizeAp);
		gd.addSlider ("FWHM multiplication factor :", 0.1, 5.1, apFWHMFactor);
        gd.addMessage (" ");
        gd.addCheckbox ("Allow left/right double click fast zoom-in/out (adds slight delay to aperture placement).", enableDoubleClicks);
		}

	/**
	 * Last part of non-required dialog created by addFancyDialog().
	 * Sub-classes not using the original dialog() will need a dummy version of this method!
	 */
	protected boolean finishFancyDialog (GenericDialog gd)
		{
		// GET NON-REQUIRED DIALOGUE FIELDS:
		//	showRatio,showRatioError,showRatioSNR,useVarSizeAp,apFWHMFactor
        reposition     = gd.getNextBoolean();
        updatePlot     = gd.getNextBoolean();
		showRatio      = gd.getNextBoolean();
        showCompTot    = gd.getNextBoolean();
		showRatioError = gd.getNextBoolean();
		showRatioSNR   = gd.getNextBoolean();
		useVarSizeAp   = gd.getNextBoolean();
		apFWHMFactor   = gd.getNextNumber();
        enableDoubleClicks = gd.getNextBoolean();
		if (gd.invalidNumber() || apFWHMFactor <= 0)
			{
            IJ.beep();
			IJ.error ("Invalid aperture FWHM factor");
            return false;
			}
		// follow = gd.getNextBoolean();
        Prefs.set (Aperture_.AP_PREFS_REPOSITION, reposition);
		Prefs.set (MultiAperture_.PREFS_SHOWRATIO, showRatio);
        Prefs.set (MultiAperture_.PREFS_SHOWCOMPTOT, showCompTot);
		Prefs.set (MultiAperture_.PREFS_SHOWRATIO_ERROR, showRatioError);
		Prefs.set (MultiAperture_.PREFS_SHOWRATIO_SNR, showRatioSNR);
		Prefs.set (MultiAperture_.PREFS_USEVARSIZEAP, useVarSizeAp);
		Prefs.set (MultiAperture_.PREFS_APFWHMFACTOR, apFWHMFactor);
        Prefs.set (MultiAperture_.PREFS_ENABLEDOUBLECLICKS, enableDoubleClicks);
        Prefs.set (MultiAperture_.PREFS_UPDATEPLOT, updatePlot);
		return true;
		}
	}

